// Interface.h:
//
// This file defines the common interfaces used by all components
// to construct and manage a model. ALthough it is perfectly possible
// to work entirely with Interfaces there is a separate class
// hierarchy that takes most of the work out of defining and
// declaring components.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_PLUGIN_H__CE49F918_3FD4_4C19_8351_829BA09C553A__INCLUDED_)
#define AFX_PLUGIN_H__CE49F918_3FD4_4C19_8351_829BA09C553A__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

// temporary until we move them
#include "Types.h"

extern "C"
{


// Common elements of all interfaces. This should be the first element
// in any interface, Interface management code relies on it. Having the
// implementation pointer as part of the interface is probably not the
// best idea but it works OK for me.
// 
// If you know what the Interface is, you can cast it to a known
// interface type - e.g. one of the ones below.
//
// Query may be a NULL pointer
struct Interface
{
	Word size;			// size of the entire interface in bytes
	const char* type;	// Interface type. Unique to each class of interface
 	void *impl;			// opaque pointer to an implementation of the interface
};

// This can be used if the normal interface conventions are adhered to
#define INTERFACE_INIT(itype, i, stype, owner)	\
	(i)->iface.size = sizeof(itype);				\
	(i)->iface.type = stype;						\
	(i)->iface.impl = owner;						\


struct IComponentClass;

// Components are the core element of the model structure. They
// represent concrete elements of the emulator such as the processor
// and memory. Each component may export a number of interfaces and
// implement others. All components must export an IComponent interface
// This interface should be type "Component"
struct IComponent
{
	Interface iface;		// generic interface
	
	IComponentClass* isa;	// optional class this component belongs to
	void* opaque;			// private data used by the component manager

	Word (*Ioctl)(IComponent*, Word op, void* in, void* out); // component ops
};

enum ComponentIoctls {
	COMPONENT_IOCTL_INIT,
	COMPONENT_IOCTL_CLEANUP,
	COMPONENT_IOCTL_CONNECT,
	COMPONENT_IOCTL_DISCONNECT,
	COMPONENT_IOCTL_START,
	COMPONENT_IOCTL_STOP,
	COMPONENT_IOCTL_RESET,
	COMPONENT_IOCTL_POSTRESET,
	COMPONENT_IOCTL_DELETE,
	COMPONENT_IOCTL_OKQUIT,
};

// Some Components need to be dynamically created at runtime from a config
// file. They do this by registering an IComponentClass interface
// which knows how to create the component.
// This interface should be type "ComponentClass"
struct IComponentClass
{
	Interface iface;	// generic interface

	const char* name;	// class name

	// Factory method for creating a component of this class
	IComponent* (*Create)(IComponentClass*);

	// used to form a list of classes
	IComponentClass *next_class;
};


// EXPERIMENTAL
// Generic property interface. Components may export one or more
// property interfaces. This currently shares a namespace with
// other interfaces.
struct IProperties
{
	Interface iface;		// generic interface

	BOOL (*SetProperty)(IProperties*, Word property, void* value);
	BOOL (*GetProperty)(IProperties*, Word property, void* value);
};


// The Component manager is responsible for overall management
// of interfaces and components. There is only one component
// manager implementation but there will typically be 
// a component manager proxy class in each compilation module.
// This interface should be type "ComponentManager"
struct IComponentManager
{
	Interface iface;	// generic interface

	// Register a component class so it can be used to instance a component
	Word (*RegisterClass)(IComponentManager*, IComponentClass*);

	// Register a public interface so other components can connect to it
	Word (*RegisterInterface)(IComponentManager*, Interface*, BOOL global);

	// Find a public interface exported by a component
	Interface* (*FindInterface)(IComponentManager*, const char* name, Word instance);

	// Find a component in the current model. This is currently only useful
	// for setting properties
	IComponent* (*FindComponent)(IComponentManager*, const char* name, Word instance);

	// Register a component so the component can be consructed properly
	Word (*AddComponent)(IComponentManager*, IComponent*, BOOL persistent);
	Word (*RemoveComponent)(IComponentManager*, IComponent*);

	// Register a global component visible in _all_ models
	Word (*AddGlobalComponent)(IComponentManager*, IComponent*);
	Word (*RemoveGlobalComponent)(IComponentManager*, IComponent*);

	// Load/store state related to a particular component (not interface)
	// Maybe we'll put these in a a generic Ioctl as there will be more at
	// a future date
	
	// If these are used with a null component, LoadString and LoadInt return
	// generic information about the application.
	void (*StoreString)(IComponentManager*, IComponent* comp, const char* key, const char* val);
	const char* (*LoadString)(IComponentManager*, IComponent* comp, const char* key, const char* deflt = "");
	void (*StoreInt)(IComponentManager*, IComponent* comp, const char* key, Word val);
	Word (*LoadInt)(IComponentManager*, IComponent* comp, const char* key, Word deflt = 0);

	// array of search paths. This is typically: plugin filter, model folter, app folder and ./
	// I don't like this so it will change
	const char** (*GetSearchPath)(IComponentManager*, Word handle);
};


// the global component manager object
extern IComponentManager* g_component_manager;

// Handle unique to each plugin module
extern Word g_plugin_handle;

} // extern "C"

#endif // AFX_PLUGIN_H__CE49F918_3FD4_4C19_8351_829BA09C553A__INCLUDED_