//////////////////////////////////////////////////////////////////////
//
// Interfaces exported by the IO controller. These are divided up as
// they may become more generic at some point
//
//////////////////////////////////////////////////////////////////////
#if !defined(IOC_INTERFACE_H)
#define IOC_INTERFACE_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Interface.h"

// names for Ioc exported interfaces

#define IFACE_IOC_IRQINTERRUPT "IRQInterrupt"
#define IFACE_IOC_FIQINTERRUPT "FIQInterrupt"
#define IFACE_IOC_DMAINTERRUPT "DMAInterrupt"
#define IFACE_IOC_CONTROLPORT  "IocControl"

// This data shouldn't be modified by the interrupt user. 
struct IInterrupt
{
	Interface iface;

	Word irq_mask;		// private data for the IOC
	Word shared_mask;	// private date for the IOC

	Word interrupt_no;	// Interrupt number this interface uses.
};

// Ioc actually exports this two or three times. Once for normal interrupts,
// once for fast interrupts and optionally for DMA interrupts. Users will need
// to request IoController instance 0, 1 or 2.

struct IIoController
{
	Interface iface;

	// Claim or release an interrupt. This will generally succeed
	BOOL (*ClaimInterrupt)(IIoController*, IInterrupt*, Word interrupt);
	BOOL (*ReleaseInterrupt)(IIoController*, IInterrupt*);

	// Set or clear an interrupt. This is fast but only use it if the device
	// was never designed to share interrupts (e.g. built in, sound or edge
	// interrupts)
	void (*SetInterrupt)(IIoController*, IInterrupt*);
	void (*ClearInterrupt)(IIoController*, IInterrupt*);

	// Set or clear an interrupt. This sets or clears only this device
	// interrupt. If there are other devices on this interrupt
	// line then their wire or'd/and'd state will be taken into account.
	void (*SetSharedInterrupt)(IIoController*, IInterrupt*);
	void (*ClearSharedInterrupt)(IIoController*, IInterrupt*);
};


// Generic Wire or'd Io Port interface.

// A device on the port
struct IIoPortDevice
{
	Interface iface;

	void (*WriteByte)(IIoPortDevice*, Byte data);
	Byte (*ReadByte)(IIoPortDevice*);
};

// bus control interface
struct IIoPort
{
	Interface iface;

	void (*AddDevice)(IIoPort*, IIoPortDevice*);
	void (*RemoveDevice)(IIoPort*, IIoPortDevice*);

	// for simple read only devices, this can set the state. But
	// this doesn't cope with shared lines very well. All devices
	// should really add/remove themselves to set state.
	void (*SetState)(IIoPort*, Byte mask, Byte state);
};



// Standard interrupt definitions for acorn machines

#define IRQA_I		7	/* Force interrupt */
#define IRQA_TM1	6	/* timer 1 */
#define IRQA_TM0	5	/* timer 0 */
#define IRQA_POR	4	/* power on reset */
#define IRQA_VFLY	3	/* video flyback */
#define IRQA_PACK	2	/* printer acknowledge */
#define IRQA_RII	1	/* serial ring indicator */
#define IRQA_PBSY	0	/* printer busy line */

// A5000 differentcs

#define IRQA_FLIX	2	/* floppy index bit */

// bits for irq status b - read only - pre A5000

#define IRQB_SRX	15	/* keyboard receive interrupt */
#define IRQB_STX	14	/* keyboard transmit interrupt */
#define IRQB_PIRQ	13	/* podule IRQ */
#define IRQB_DCIRQ	12	/* disc change */
#define IRQB_WIRQ	11	/* hard disc interrupt */
#define IRQB_SLCI	10	/* Serial controller */
#define IRQB_SIRQ	9	/* sound buffer reload */
#define IRQB_PFIQ	8	/* Podule FIQ */

// post a5000 bits

#define IRQB_IDE	11	/* ide interrupt */
#define IRQB_FLOP	12	/* floppy interrupt */

// post riscpc

#define IRQB_RPCIDE 9  /* riscpc ide interrupt */
#define IRQB_RPCNET 11	/* riscpc network card */

// bits for fiq_status - read only - pre A5000

#define FIQ_I		7	/* force fiq */
#define FIQ_PFIQ	6	/* Podulq FIQ */
#define FIQ_C5		5	/* pin C5 status */
#define FIQ_C4		4	/* pin C4 status */
#define FIQ_C3		3	/* pin C3 status */
#define FIQ_EFIQ	2	/* econet */
#define FIQ_FFIQ	1	/* floppy interrupt */
#define FIQ_FFDQ	0	/* floppy data interrupt */


// bits for control port - read/write

#define CTRL_VFLY	0x80	/* vfly */
#define CTRL_PACK	0x40	/* printer acknowledge */
#define CTRL_SMUTE	0x20	/* sound mute */
#define CTRL_AUX	0x10	/* to aux connector */
#define CTRL_C3		0x08	/* reserved */
#define CTRL_FDR	0x04	/* floppy read */
#define CTRL_SCL	0x02	/* i2c clock */
#define CTRL_SDA	0x01	/* i2c data */

// a5000 version

#define CTRL_FLIX	0x40	/* floppy index status */

#endif // IOC_INTERFACE_H